/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.controller;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.text.MessageFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.servlet.http.HttpSession;

import org.apache.commons.lang3.text.WordUtils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.servlet.ModelAndView;

import com.enerallies.weather.entity.DegreeDayData;
import com.enerallies.weather.entity.DegreeDaySearchCriteria;
import com.enerallies.weather.entity.DegreeDayType;
import com.enerallies.weather.entity.DegreeType;
import com.enerallies.weather.entity.Setpoint;
import com.enerallies.weather.entity.SetpointSchedule;
import com.enerallies.weather.entity.WeatherData;
import com.enerallies.weather.entity.WeatherDataSearchCriteria;
import com.enerallies.weather.service.WeatherDataService;
import com.enerallies.weather.service.WeatherDataServiceException;

/**
 * Controller for the weather data pages.
 *
 * <p>
 * <b>Thread Safety:</b> This class is not thread safe since it is mutable but
 * it will be used in a thread safe manner my Spring framework.
 * </p>
 *
 * @author Standlove, hanno
 * @version 1.0
 */
@Controller
public class WeatherDataController {
    /**
     * Class logger.
     */
    private static final Logger LOGGER = Logger.getLogger(WeatherDataController.class);

    /**
     * Week days.
     */
    private static final String[] DAYS = {"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };

    /**
     * Prefix weather data.
     */
    private static final String PREFIX_WEATHER_DATA = "searchWeatherData";

    /**
     * Prefix degree day data.
     */
    private static final String PREFIX_DEGREE_DAY = "searchDegreeDayData";

    /**
     * Session key for the form object.
     */
    private static final String SESSION_KEY_WD_FORM = "wd_form";

    /**
     * Session key for the form object.
     */
    private static final String SESSION_KEY_DD_FORM = "dd_form";

    /**
     * Comma.
     */
    private static final String COMMA = ";";

    /**
     * The injected service.
     */
    @Autowired
    private WeatherDataService service;

    /**
     * Empty constructor.
     */
    public WeatherDataController() {
        // empty
    }

    /**
     * GET index.
     *
     * @return view for index
     */
    @RequestMapping(value = {"/", "/index" }, method = RequestMethod.GET)
    public String index() {
        return "index";
    }

    /**
     * GET searchWeatherData.
     *
     * @param session
     *            HTTP session
     * @return view for searchWeatherData
     */
    @RequestMapping(value = "/searchWeatherData", method = RequestMethod.GET)
    public ModelAndView searchWeatherData(HttpSession session) {
        Object form = session.getAttribute(SESSION_KEY_WD_FORM);
        if (form == null) {
            form = new SearchWeatherDataForm();
        }
        return new ModelAndView("searchWeatherData", "form", form);
    }

    /**
     * This method is called when the user clicks the search button on the
     * weather data input screen.
     * <p>
     * It will call service.retrieveWeatherData method to get the weather data.
     * The criteria and retrieved weatherData should be saved to session (to be
     * used in exportWeatherData method).
     * </p>
     *
     * The weatherData will be rendered in weather data output screen.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return model and view for the weather data output screen
     * @throws WeatherDataServiceException
     *             if any error occur
     */
    @RequestMapping(value = "/searchWeatherData", method = RequestMethod.POST)
    public ModelAndView searchWeatherData(@ModelAttribute("form") SearchWeatherDataForm form, HttpSession session)
        throws WeatherDataServiceException {
        try {
            session.setAttribute(SESSION_KEY_WD_FORM, form);
            return searchWeatherDataWithIAE(form, session);
        } catch (WeatherDataServiceException e) {
            LOGGER.debug("Could not perform search: " + e.getMessage(), e);
            throw e;
        }
    }

    /**
     * Search with all exceptions caught.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return model and view for the weather data output screen
     * @throws WeatherDataServiceException
     *             if any error occur
     */
    private ModelAndView searchWeatherDataWithIAE(SearchWeatherDataForm form, HttpSession session)
        throws WeatherDataServiceException {
        try {
            Map<Date, List<WeatherData>> data = searchWeatherDataInternal(form, session);

            ModelAndView mv = new ModelAndView("showWeatherData", "weatherData", data);

            // add the form to fill hidden input in JSP
            mv.addObject("locationOrZip", form.getLocationOrZip());
            mv.addObject("beginningDate", form.getBeginningDate());
            mv.addObject("endDate", form.getEndDate());

            return mv;
        } catch (IllegalArgumentException e) {
            throw new WeatherDataServiceException("Invalid values in input form: " + e.getMessage(), e);
        }
    }

    /**
     * Filters values in the future.
     *
     * @param data
     *            data
     * @return filtered map (without values in the future)
     */
    private Map<Date, List<WeatherData>> filterFutureValues(Map<Date, List<WeatherData>> data) {
        Map<Date, List<WeatherData>> filtered = new HashMap<Date, List<WeatherData>>();

        Date now = new Date();
        for (Entry<Date, List<WeatherData>> e : data.entrySet()) {
            List<WeatherData> newWds = new ArrayList<WeatherData>();
            for (WeatherData wd : e.getValue()) {
                if (wd.getUtcDate().before(now)) {
                    newWds.add(wd);
                }
            }

            if (e.getKey().before(now)) {
                filtered.put(e.getKey(), newWds);
            }
        }

        return filtered;
    }

    /**
     * Perform the actual search and put result into session.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return the search result
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    private Map<Date, List<WeatherData>> searchWeatherDataInternal(SearchWeatherDataForm form, HttpSession session)
        throws WeatherDataServiceException {
        WeatherDataSearchCriteria criteria = new WeatherDataSearchCriteria();
        initCriteria(criteria, form);

        Map<Date, List<WeatherData>> data = filterFutureValues(this.service.retrieveWeatherData(criteria));
        session.setAttribute(createFileName(PREFIX_WEATHER_DATA, form), data);
        return data;
    }

    /**
     * Fill the values from the criteria.
     *
     * @param criteria
     *            the criteria
     * @param form
     *            the input form
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    private void initCriteria(WeatherDataSearchCriteria criteria, SearchWeatherDataForm form)
        throws WeatherDataServiceException {
        if (form.getLocationOrZip() == null) {
            throw new WeatherDataServiceException("Please insert location or ZIP in the search field.");
        }

        criteria.setBeginningDate(parseDate(form.getBeginningDate()));
        criteria.setEndDate(parseDate(form.getEndDate()));

        if (form.getLocationOrZip().matches("[0-9]*")) {
            criteria.setZipCode(form.getLocationOrZip());
        } else {
            String[] splitted = form.getLocationOrZip().split(",");
            if (splitted.length == 2) {
                criteria.setCity(WordUtils.capitalize(splitted[0].trim()));
                criteria.setState(WordUtils.capitalize(splitted[1].trim()));
                form.setLocationOrZip(criteria.getCity() + ", " + criteria.getState());
            } else {
                throw new WeatherDataServiceException("Invalid search criteria: " + form.getLocationOrZip());
            }
        }
    }

    /**
     * Fill the values from the criteria.
     *
     * @param criteria
     *            the criteria
     * @param form
     *            the input form
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    private void initCriteria(DegreeDaySearchCriteria criteria, SearchDegreeDayForm form)
        throws WeatherDataServiceException {
        initCriteria((WeatherDataSearchCriteria) criteria, form);

        criteria.setDegreeDayType(DegreeDayType.valueOf(form.getDegreeDayType()));
        criteria.setDegreeType(DegreeType.valueOf(form.getDegreeType()));
        criteria.setSetpointSchedule(SetpointSchedule.valueOf(form.getSetpointSchedule()));

        int start;
        int end;

        if (SetpointSchedule.valueOf(form.getSetpointSchedule()) == SetpointSchedule.SINGLE_DAY) {
            start = 0;
            end = SearchDegreeDayForm.START_OF_WEEK_SCHEDULE;
        } else {
            start = SearchDegreeDayForm.START_OF_WEEK_SCHEDULE;
            end = form.getSetpoints().length;
        }

        SimpleDateFormat df = new SimpleDateFormat("HH:mm");

        Setpoint[] setpoints = new Setpoint[end - start];
        for (int i = start; i < end; i++) {
            String spString = form.getSetpoints()[i];
            String timeString = form.getTimes()[i];

            if (spString == null || spString.isEmpty()) {
                continue;
            }

            Setpoint sp = new Setpoint();
            setpoints[i - start] = sp;

            try {
                sp.setSetpoint(Double.valueOf(spString).doubleValue());
            } catch (NumberFormatException e) {
                throw new WeatherDataServiceException("Invalid setpoint in " + getScheduleDescription(i), e);
            }

            if (timeString == null || timeString.isEmpty()) {
                throw new WeatherDataServiceException("Missing time in " + getScheduleDescription(i));
            }

            try {
                sp.setTime(df.parse(timeString));
            } catch (ParseException e) {
                throw new WeatherDataServiceException("Invalid time in " + getScheduleDescription(i), e);
            }
        }

        criteria.setSetpoints(Arrays.asList(setpoints));
    }

    /**
     * Returns the description for the schedule value at position i.
     *
     * @param i
     *            position
     * @return the description
     */
    private String getScheduleDescription(int i) {
        if (i < SearchDegreeDayForm.START_OF_WEEK_SCHEDULE) {
            return "single day schedule row " + (i + 1);
        }

        return "7 days schedule row " + ((i % SearchDegreeDayForm.START_OF_WEEK_SCHEDULE) + 1) + " on "
            + DAYS[(i / SearchDegreeDayForm.START_OF_WEEK_SCHEDULE) - 1];
    }

    /**
     * Create a session key (file name).
     *
     * @param prefix
     *            a prefix
     * @param form
     *            the input form
     * @return session key (file name)
     */
    private String createFileName(String prefix, SearchWeatherDataForm form) {
        return MessageFormat.format(prefix + "({0},{1},{2})", form.getLocationOrZip(), form.getBeginningDate(),
            form.getEndDate());
    }

    /**
     * Create a session key starting with the file name followed by additional
     * values from the search criteria.
     *
     * @param form
     *            the input form
     * @return session key
     */
    private String createSessionKey(SearchDegreeDayForm form) {
        StringBuilder b = new StringBuilder(createFileName(PREFIX_DEGREE_DAY, form));

        b.append(form.getDegreeDayType()).append(COMMA);
        b.append(form.getDegreeType()).append(COMMA);
        b.append(form.getSetpointSchedule()).append(COMMA);
        b.append(Arrays.asList(form.getTimes()));
        b.append(Arrays.asList(form.getSetpoints())).append(COMMA);

        return b.toString();
    }

    /**
     * GET searchWeatherData.
     *
     * @param session
     *            HTTP session
     * @return view for searchWeatherData
     */
    @RequestMapping(value = "/searchDegreeDayData", method = RequestMethod.GET)
    public ModelAndView searchDegreeDayData(HttpSession session) {
        Object form = session.getAttribute(SESSION_KEY_DD_FORM);
        if (form == null) {
            form = new SearchDegreeDayForm();
        }
        return new ModelAndView("searchDegreeDayData", "form", form);
    }

    /**
     * This method is called when the user clicks the search button on the
     * degree day input screen.
     * <p>
     * It will call service.retrieveDegreeDayData method to get the weather
     * data. The criteria and retrieved degreeDayData should be saved to session
     * (to be used in exportDegreeDayData method).
     * </p>
     *
     * The degreeDayData will be rendered in degree day data output screen.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return model and view for the degree day data output screen
     * @throws WeatherDataServiceException
     *             if any error occur
     */
    @RequestMapping(value = "/searchDegreeDayData", method = RequestMethod.POST)
    public ModelAndView searchDegreeDayData(@ModelAttribute("form") SearchDegreeDayForm form, HttpSession session)
        throws WeatherDataServiceException {
        try {
            session.setAttribute(SESSION_KEY_DD_FORM, form);
            return searchDegreeDayDataWithIAE(form, session);
        } catch (WeatherDataServiceException e) {
            LOGGER.debug("Could not perform search: " + e.getMessage(), e);
            throw e;
        }
    }

    /**
     * Search with all exceptions caught.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return model and view for the degree day data output screen
     * @throws WeatherDataServiceException
     *             if any error occur
     */
    private ModelAndView searchDegreeDayDataWithIAE(SearchDegreeDayForm form, HttpSession session)
        throws WeatherDataServiceException {
        try {
            DegreeDayData data = searchDegreeDayDataInternal(form, session);

            ModelAndView mv = new ModelAndView("showDegreeDayData", "degreeDayData", data);

            // add the form to fill hidden input in JSP
            mv.addObject("locationOrZip", form.getLocationOrZip());
            mv.addObject("beginningDate", form.getBeginningDate());
            mv.addObject("endDate", form.getEndDate());
            mv.addObject("degreeType", form.getDegreeType());
            mv.addObject("degreeDayType", form.getDegreeDayType());
            mv.addObject("setpoints", form.getSetpoints());
            mv.addObject("setpointSchedule", form.getSetpointSchedule());
            mv.addObject("times", form.getTimes());

            return mv;
        } catch (IllegalArgumentException e) {
            throw new WeatherDataServiceException("Invalid values in input form: " + e.getMessage(), e);
        }
    }

    /**
     * Perform the actual search and put result into session.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return the search result
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    private DegreeDayData searchDegreeDayDataInternal(SearchDegreeDayForm form, HttpSession session)
        throws WeatherDataServiceException {
        DegreeDaySearchCriteria criteria = new DegreeDaySearchCriteria();
        initCriteria(criteria, form);

        DegreeDayData data = this.service.retrieveDegreeDayData(criteria);
        session.setAttribute(createSessionKey(form), data);

        return data;
    }

    /**
     * This method is called when the export data vertically (or horizontally)
     * buttons are clicked on the weather data output screen.
     * <p>
     * It will check if the criteria is the same as the one in session or not,
     * if yes, retrieve the weatherData from session. Otherwise, call
     * service.retrieveWeatherData to retrieve the weather data.
     * </p>
     * Calls service.exportWeatherData to export the data into response's output
     * stream. The content-type & content-disposition of the response will be
     * set properly.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return the csv data in an HTTP entity
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    @RequestMapping(value = "/exportWeatherData", method = RequestMethod.POST)
    public ResponseEntity<byte[]> exportWeatherData(@ModelAttribute("form") SearchWeatherDataForm form,
        HttpSession session) throws WeatherDataServiceException {
        try {
            return exportWeatherDataWithIAE(form, session);
        } catch (WeatherDataServiceException e) {
            LOGGER.debug("Could not perform export: " + e.getMessage(), e);
            throw e;
        }
    }

    /**
     * Export with all exception caught.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return the csv data in an HTTP entity
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    private ResponseEntity<byte[]> exportWeatherDataWithIAE(SearchWeatherDataForm form, HttpSession session)
        throws WeatherDataServiceException {
        try {
            String sessionKey = createFileName(PREFIX_WEATHER_DATA, form);
            @SuppressWarnings("unchecked")
            Map<Date, List<WeatherData>> data = (Map<Date, List<WeatherData>>) session.getAttribute(sessionKey);

            if (data == null) {
                // nothing found in session, do a query
                data = searchWeatherDataInternal(form, session);
            }

            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            this.service.exportWeatherData(data, form.getExportDirection(), bout);

            return createCsvEntity(bout, sessionKey);
        } catch (IllegalArgumentException e) {
            throw new WeatherDataServiceException("Invalid values in input form: " + e.getMessage(), e);
        }
    }

    /**
     * Create a CSV entity.
     *
     * @param bout
     *            the data
     * @param name
     *            the name
     * @return a CSV entity
     */
    private ResponseEntity<byte[]> createCsvEntity(ByteArrayOutputStream bout, String name) {
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_OCTET_STREAM);
        headers.setContentDispositionFormData("attachment", name + ".csv");

        return new ResponseEntity<byte[]>(bout.toByteArray(), headers, HttpStatus.OK);
    }

    /**
     * It is called when the export data vertically (or horizontally) buttons
     * are clicked on the degree day data output screen.
     * <p>
     * It will check the criteria is the same as the one in session or not, if
     * yes, retrieve the degreeDayData from session. Otherwise, call
     * service.retrieveDegreeDayData to retrieve the degree day data.
     * </p>
     * Calls service.exportDegreeDayData to export the data into response's
     * output stream. The content-type & content-disposition of the response
     * will be set properly.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return CSV data
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    @RequestMapping(value = "/exportDegreeDayData", method = RequestMethod.POST)
    public ResponseEntity<byte[]> exportDegreeDayData(@ModelAttribute("form") SearchDegreeDayForm form,
        HttpSession session) throws WeatherDataServiceException {
        try {
            return exportDegreeDayDataWithIAE(form, session);
        } catch (WeatherDataServiceException e) {
            LOGGER.debug("Could not perform export: " + e.getMessage(), e);
            throw e;
        }
    }

    /**
     * Export with all exceptions caught.
     *
     * @param form
     *            the input form
     * @param session
     *            HTTP session
     * @return CSV data
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    private ResponseEntity<byte[]> exportDegreeDayDataWithIAE(SearchDegreeDayForm form, HttpSession session)
        throws WeatherDataServiceException {
        try {
            String sessionKey = createSessionKey(form);
            DegreeDayData data = (DegreeDayData) session.getAttribute(sessionKey);

            if (data == null) {
                // nothing found in session, do a query
                data = searchDegreeDayDataInternal(form, session);
            }

            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            this.service.exportDegreeDayData(data, form.getExportDirection(), bout);

            return createCsvEntity(bout, createFileName(PREFIX_DEGREE_DAY, form));
        } catch (IllegalArgumentException e) {
            throw new WeatherDataServiceException("Invalid values in input form: " + e.getMessage(), e);
        }
    }

    /**
     * Sets the weather service.
     *
     * @param service
     *            the new weather service to set
     */
    public void setService(WeatherDataService service) {
        this.service = service;
    }

    /**
     * Parse given date.
     *
     * @param d
     *            date to be parsed
     * @return parsed date
     * @throws WeatherDataServiceException
     *             if it can not be parsed
     */
    private Date parseDate(String d) throws WeatherDataServiceException {
        if (d == null) {
            return null;
        }

        try {
            return new SimpleDateFormat("MM/dd/yyyy").parse(d);
        } catch (ParseException e) {
            throw new WeatherDataServiceException("Could not parse date value: " + e.getMessage(), e);
        }
    }

    /**
     * This method is called when the user clicks the export button on the
     * degree day input screen.
     * <p>
     * It will export the setpoint schedule currently filled in into a
     * properties file.
     * </p>
     *
     *
     * @param form
     *            the input form
     * @return exported data
     */
    @RequestMapping(value = "/exportSetpointSchedule", method = RequestMethod.POST)
    public ResponseEntity<byte[]> exportSetpointSchedule(@ModelAttribute("form") SearchDegreeDayForm form) {
        PrintWriter pw = null;
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        try {
            pw = new PrintWriter(new BufferedOutputStream(bout));

            pw.println(form.getSetpointSchedule());
            for (int i = 0; i < SearchDegreeDayForm.MAX_SETPOINTS; i++) {
                String time = form.getTimes()[i];
                if (time == null || time.trim().isEmpty()) {
                    time = "HH:MM";
                }
                pw.println(time + "," + form.getSetpoints()[i]);
            }
        } finally {
            if (pw != null) {
                pw.close();
            }
        }

        return createCsvEntity(bout, "setpointSchedule");
    }

    /**
     * This method is called when the user clicks the import button on the
     * degree day input screen.
     *
     * @param session
     *            HTTP session
     * @return model and view
     */
    @RequestMapping(value = "/importSetpointSchedule", method = RequestMethod.GET)
    public ModelAndView importSetpointSchedule(HttpSession session) {
        return searchDegreeDayData(session);
    }

    /**
     * This method is called when the user clicks the import button on the
     * degree day input screen.
     *
     * @param session
     *            HTTP session
     *
     * @param form
     *            the input form
     * @return model and view
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    @RequestMapping(value = "/importSetpointSchedule", method = RequestMethod.POST)
    public ModelAndView importSetpointSchedule(HttpSession session, @ModelAttribute("form") SearchDegreeDayForm form)
        throws WeatherDataServiceException {

        session.setAttribute(SESSION_KEY_DD_FORM, form);

        if (form.getImportFile() == null) {
            throw new WeatherDataServiceException("No import file found");
        }

        BufferedReader r = null;
        try {
            r = new BufferedReader(new InputStreamReader(form.getImportFile().getInputStream()));
            form.setSetpointSchedule(r.readLine());

            String line = r.readLine();
            int i = 0;
            while (line != null && i < SearchDegreeDayForm.MAX_SETPOINTS) {
                String[] splitted = line.split(",");
                if (splitted.length == 1) {
                    form.getTimes()[i] = splitted[0];
                    form.getSetpoints()[i] = "";
                } else if (splitted.length == 2) {
                    form.getTimes()[i] = splitted[0];
                    form.getSetpoints()[i] = splitted[1];
                } else {
                    throw new WeatherDataServiceException("Invalid line in file: " + line);
                }

                line = r.readLine();
                i++;
            }
        } catch (IOException e) {
            throw new WeatherDataServiceException("Could not read import file: " + e.getMessage(), e);
        } finally {
            if (r != null) {
                try {
                    r.close();
                } catch (IOException e) {
                    throw new WeatherDataServiceException("Could not read import file: " + e.getMessage(), e);
                }
            }
        }

        return new ModelAndView("searchDegreeDayData", "form", form);
    }
}
