/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.service;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import com.enerallies.weather.entity.WeatherData;
import com.enerallies.weather.entity.WeatherDataSearchCriteria;
import com.enerallies.weather.service.Response.Observation;
import com.enerallies.weather.service.Response.ResponseDate;

/**
 * Contains a list of history query result.
 *
 * @author hanno
 * @version 1.0
 */
class HistoryQueryResultList {
    /**
     * The number of weather data entries per day.
     */
    private static final int ENTRIES_PER_DAY = 24;

    /**
     * 50 percent.
     */
    private static final double PERCENT_50 = 0.5;

    /**
     * The internal data list.
     */
    private final List<HistoryQueryResult> data = new ArrayList<HistoryQueryResult>();

    /**
     * The original response observations.
     */
    private final List<Observation> responseObservations = new ArrayList<Observation>();

    /**
     * The search criteria.
     */
    private final WeatherDataSearchCriteria criteria;

    /**
     * Create a new instance.
     *
     * @param criteria
     *            criteria
     */
    public HistoryQueryResultList(WeatherDataSearchCriteria criteria) {
        this.criteria = criteria;
    }

    /**
     * Add a new response for the given date.
     *
     * @param date
     *            date for which the query was made
     * @param response
     *            the response
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    public void add(Calendar date, Response response) throws WeatherDataServiceException {
        data.add(new HistoryQueryResult(date, response));
    }

    /**
     * Returns the data.
     *
     * @return the data.
     */
    public List<HistoryQueryResult> getData() {
        return data;
    }

    /**
     * To map.
     *
     * @return this last of map date -> list
     * @throws WeatherDataServiceException
     *             if an error occurs
     */
    public Map<Date, List<WeatherData>> toMap() throws WeatherDataServiceException {
        Map<Date, List<WeatherData>> result = new HashMap<Date, List<WeatherData>>();

        for (HistoryQueryResult hq : data) {
            List<WeatherData> weatherList = new ArrayList<WeatherData>();
            for (Observation o : hq.observations) {
                WeatherData d = o.toWeatherData();
                d.setCriteria(criteria);
                weatherList.add(d);
            }
            result.put(hq.date.getTime(), weatherList);
        }

        return result;
    }

    /**
     * This method will interpolate 24 values for each day searched for using
     * linear interpolation. It will:
     * <ul>
     * <li>Accumulate all Observations from all responses in a single sorted
     * list</li>
     * <li>Starting from the first day at 00:00 it will calculate the values at
     * full hours (00:00, 1:00, 2:00,...) for all days from the adjacent values
     * in the sorted Observations list</li>
     * </ul>
     *
     * @throws WeatherDataServiceException
     *             if any error occurs
     */
    public void interpolate() throws WeatherDataServiceException {
        if (this.responseObservations.isEmpty()) {
            throw new WeatherDataServiceException("Query did not return any data");
        }

        Collections.sort(this.responseObservations, new ObservationDateComparator());

        int nData = data.size() * ENTRIES_PER_DAY;
        final Calendar currentDate = (Calendar) data.get(0).getDate().clone();
        currentDate.setTimeZone(this.responseObservations.get(0).getDate().toDate().getTimeZone());

        prepare(this.responseObservations);

        // make sure we start at beginning of day
        currentDate.set(Calendar.HOUR_OF_DAY, 0);
        currentDate.set(Calendar.MINUTE, 0);
        currentDate.set(Calendar.SECOND, 0);
        currentDate.set(Calendar.MILLISECOND, 0);

        Observation lookup = new Observation();
        lookup.setDate(new ResponseDate() {
            @Override
            public Calendar toDate() throws WeatherDataServiceException {
                return currentDate;
            }
        });

        int currentObservationIndex = 0;
        Observation before = this.responseObservations.get(currentObservationIndex);
        Observation after = before;
        for (int i = 0; i < nData; i++) {
            while (currentObservationIndex < this.responseObservations.size() - 1
                && after.getDate().toDate().before(currentDate)) {
                currentObservationIndex++;
                if (currentObservationIndex == this.responseObservations.size()) {
                    // reached end of list
                    before = this.responseObservations.get(currentObservationIndex - 1);
                    after = before;
                } else {
                    before = after;
                    after = this.responseObservations.get(currentObservationIndex);
                }
            }

            setDataAt(i, createDataBetween(before, after, currentDate));

            // go to the next hour
            currentDate.add(Calendar.HOUR_OF_DAY, 1);
        }

    }

    /**
     * Prepare the list of observations by filling missing values in single
     * observations.
     *
     * @param responseObservations
     *            list of observations.
     */
    private void prepare(List<Observation> responseObservations) {
        Double temperature = null;
        Double dewPoint = null;
        Integer humidity = null;
        Double visibility = null;
        String windDirection = null;
        Double windSpeed = null;
        Double gustSpeed = null;
        Double precip = null;

        for (Observation o : responseObservations) {
            if (temperature == null && o.getTemperature() != null) {
                temperature = o.getTemperature();
            }
            if (dewPoint == null && o.getDewPoint() != null) {
                dewPoint = o.getDewPoint();
            }
            if (humidity == null && o.getHumidity() != null) {
                humidity = o.getHumidity();
            }
            if (visibility == null && o.getVisibility() != null) {
                visibility = o.getVisibility();
            }
            if (windDirection == null && o.getWindDirection() != null) {
                windDirection = o.getWindDirection();
            }
            if (windSpeed == null && o.getWindSpeed() != null) {
                windSpeed = o.getWindSpeed();
            }
            if (gustSpeed == null && o.getGustSpeed() != null) {
                gustSpeed = o.getGustSpeed();
            }
            if (precip == null && o.getPrecip() != null) {
                precip = o.getPrecip();
            }
        }

        for (Observation o : responseObservations) {
            if (o.getTemperature() == null) {
                o.setTemperature(temperature);
            } else {
                temperature = o.getTemperature();
            }
            if (o.getDewPoint() == null) {
                o.setDewPoint(dewPoint);
            } else {
                dewPoint = o.getDewPoint();
            }
            if (o.getHumidity() == null) {
                o.setHumidity(humidity);
            } else {
                humidity = o.getHumidity();
            }
            if (o.getVisibility() == null) {
                o.setVisibility(visibility);
            } else {
                visibility = o.getVisibility();
            }
            if (o.getWindDirection() == null) {
                o.setWindDirection(windDirection);
            } else {
                windDirection = o.getWindDirection();
            }
            if (o.getWindSpeed() == null) {
                o.setWindSpeed(windSpeed);
            } else {
                windSpeed = o.getWindSpeed();
            }
            if (o.getGustSpeed() == null) {
                o.setGustSpeed(gustSpeed);
            } else {
                gustSpeed = o.getGustSpeed();
            }
            if (o.getPrecip() == null) {
                o.setPrecip(precip);
            } else {
                precip = o.getPrecip();
            }
        }
    }

    /**
     * Create interpolated weather data.
     *
     * @param d1
     *            first value (may be null)
     * @param d2
     *            second value (may be null);
     * @param date
     *            the date for the result
     * @return interpolated weather data.
     * @throws WeatherDataServiceException
     *             if an error occurs
     */
    private Observation createDataBetween(Observation d1, Observation d2, Calendar date)
        throws WeatherDataServiceException {

        Observation d = d1.clone();
        double percent;
        double diff = d2.getDate().toDate().getTimeInMillis() - d1.getDate().toDate().getTimeInMillis();
        if (diff != 0) {
            percent = (date.getTimeInMillis() - d1.getDate().toDate().getTimeInMillis()) / diff;
        } else {
            // no difference means exact match -> 100%
            percent = 1;
        }
        percent = Math.max(0, Math.min(1, percent));

        d.setDewPoint(interpolated(d1.getDewPoint(), d2.getDewPoint(), percent));
        d.setGustSpeed(interpolated(d1.getGustSpeed(), d2.getGustSpeed(), percent));
        d.setHumidity(interpolated(d1.getHumidity(), d2.getHumidity(), percent));
        d.setPrecip(interpolated(d1.getPrecip(), d2.getPrecip(), percent));
        d.setTemperature(interpolated(d1.getTemperature(), d2.getTemperature(), percent));
        d.setVisibility(interpolated(d1.getVisibility(), d2.getVisibility(), percent));
        d.setWindDirection(percent < PERCENT_50 ? d1.getWindDirection() : d2.getWindDirection());
        d.setWindSpeed(interpolated(d1.getWindSpeed(), d2.getWindSpeed(), percent));

        d.getDate().setDate(date);

        Calendar utcDate = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        utcDate.setTime(date.getTime());
        d.getUtcDate().setDate(utcDate);

        return d;
    }

    /**
     * Interpolate integers.
     *
     * @param d1
     *            first integer
     * @param d2
     *            second integer
     * @param percent
     *            percentage
     * @return interpolated integers
     */
    private Integer interpolated(Integer d1, Integer d2, double percent) {
        if (d1 == null || d2 == null) {
            return percent < PERCENT_50 ? d1 : d2;
        }
        return Integer
            .valueOf(interpolated(Double.valueOf(d1.doubleValue()), Double.valueOf(d2.doubleValue()), percent)
                .intValue());
    }

    /**
     * Interpolate doubles.
     *
     * @param d1
     *            first double
     * @param d2
     *            second double
     * @param percent
     *            percentage
     * @return interpolated doubles
     */
    private Double interpolated(Double d1, Double d2, double percent) {
        if (d1 == null || d2 == null) {
            return percent < PERCENT_50 ? d1 : d2;
        }
        return Double.valueOf(d1.doubleValue() + (d2.doubleValue() - d1.doubleValue()) * percent);
    }

    /**
     * Returns the weather data at the given index as if all weather data in the
     * queries are in a single list.
     *
     * @param i
     *            index
     * @param d
     *            data to set
     */
    private void setDataAt(int i, Observation d) {
        Observation[] weatherDatas = data.get(i / ENTRIES_PER_DAY).observations;
        weatherDatas[i % ENTRIES_PER_DAY] = d;
    }

    /**
     * A history query for a single day.
     *
     * @author hanno
     * @version 1.0
     */
    public class HistoryQueryResult {
        /**
         * The date.
         */
        private final Calendar date;

        /**
         * All weather data for the day.
         */
        private final Observation[] observations;

        /**
         * Create a new instance.
         *
         * @param date
         *            the date
         * @param response
         *            the response
         * @throws WeatherDataServiceException
         *             if an error occurs
         */
        public HistoryQueryResult(Calendar date, Response response) throws WeatherDataServiceException {
            this.date = date;
            this.observations = new Observation[ENTRIES_PER_DAY];

            if (response.getHistory() != null) {
                responseObservations.addAll(response.getHistory().getObservations());
            }
        }

        /**
         * Returns the date.
         *
         * @return the date
         */
        public Calendar getDate() {
            return date;
        }

        /**
         * Returns the observations.
         *
         * @return the observations
         */
        public Observation[] getObservations() {
            return observations;
        }
    }

    /**
     * Comparator for Observations.
     *
     * @author hanno
     * @version 1.0
     */
    private static class ObservationDateComparator implements Comparator<Observation> {
        @Override
        public int compare(Observation o1, Observation o2) {
            try {
                return o1.getDate().toDate().compareTo(o1.getDate().toDate());
            } catch (WeatherDataServiceException e) {
                return 0;
            }
        }
    }
}
