/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.service;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.log4j.Logger;

import com.enerallies.weather.entity.DegreeDaySearchCriteria;
import com.enerallies.weather.entity.Setpoint;
import com.enerallies.weather.entity.SetpointSchedule;

/**
 * Handles access to Setpoints created from a given criteria.
 *
 * @author hanno
 * @version 1.0
 */
class SetpointList {
    /**
     * Class logger.
     */
    private static final Logger LOGGER = Logger.getLogger(SetpointList.class);

    /**
     * Days per week.
     */
    private static final int SETPOINTS_PER_DAY = 7;

    /**
     * Name of all days. Used for error messages.
     */
    private static final String[] DAYS = {"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };

    /**
     * Type SINGLE_DAY yes/no.
     */
    private final boolean isSingleDay;

    /**
     * The current day of the week.
     */
    private int dayOfWeek;

    /**
     * The list of setpoints sorted per day.
     */
    private final List<List<Setpoint>> setpoints = new ArrayList<List<Setpoint>>();

    /**
     * Creates a new instance from the given criteria.
     *
     * @param criteria
     *            the search criteria
     * @throws IllegalArgumentException
     *             if any day contains no setpoint
     */
    public SetpointList(DegreeDaySearchCriteria criteria) {
        isSingleDay = criteria.getSetpointSchedule() == SetpointSchedule.SINGLE_DAY;

        List<Setpoint> dailySetpoints = null;
        for (int i = 0; i <= criteria.getSetpoints().size(); i++) {
            if (i % SETPOINTS_PER_DAY == 0) {
                // switch to next day
                sortList(dailySetpoints);
                dailySetpoints = new ArrayList<Setpoint>();
                setpoints.add(dailySetpoints);
            }
            if (i < criteria.getSetpoints().size()) {
                dailySetpoints.add(criteria.getSetpoints().get(i));
            }
        }

        if (!hasValues()) {
            throw new IllegalArgumentException("Not a single setpoint found");
        }

        if (LOGGER.isDebugEnabled()) {
            logList();
        }
    }

    /**
     * Log the list.
     */
    private void logList() {
        SimpleDateFormat df = new SimpleDateFormat("HH:mm");
        int i = 0;
        for (List<Setpoint> ds : setpoints) {
            for (Setpoint sp : ds) {
                String msg = (i < 10 ? "0" : "") + i + ": ";
                i++;

                if (sp != null) {
                    msg += df.format(sp.getTime()) + " " + sp.getSetpoint();

                }

                LOGGER.debug(msg);
            }
        }
    }

    /**
     * Sorts the list according to the setpoint's time.
     *
     * @param dailySetpoints
     *            list to be sorted
     */
    private void sortList(List<Setpoint> dailySetpoints) {
        if (dailySetpoints != null) {
            Collections.sort(dailySetpoints, new Comparator<Setpoint>() {
                @Override
                public int compare(Setpoint o1, Setpoint o2) {
                    if (o1 == null) {
                        if (o2 == null) {
                            return 0;
                        }
                        return 1;
                    } else if (o2 == null) {
                        return -1;
                    }

                    return o1.getTime().compareTo(o2.getTime());
                }
            });
        }
    }

    /**
     * Checks if there is any value available.
     *
     * @return true if there is any value available.
     */
    private boolean hasValues() {
        for (List<Setpoint> ds : setpoints) {
            for (Setpoint sp : ds) {
                if (sp != null) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Gets the setpoint value for the given hour of day.
     *
     * @param hourOfDay
     *            hour of day
     * @return the setpoint value for the given hour of day.
     */
    public double getValue(int hourOfDay) {
        Double currentValue = null;
        for (Setpoint sp : setpoints.get(dayOfWeek)) {
            if (isHourInSetpoint(hourOfDay, sp)) {
                currentValue = Double.valueOf(sp.getSetpoint());
            } else {
                if (currentValue == null) {
                    // setpoint is in the future and we do not have a value yet
                    // example: first setpoint of day is 8:00 and hour of day is
                    // 00:00 --> last active setpoint from previous non-empty
                    // day needed
                    for (int i = 1; i <= setpoints.size(); i++) {
                        int previousDayOfWeek = ((dayOfWeek - i) + setpoints.size()) % setpoints.size();
                        List<Setpoint> previousDayList = setpoints.get(previousDayOfWeek);
                        for (int j = previousDayList.size() - 1; j >= 0; j--) {
                            if (previousDayList.get(j) != null) {
                                currentValue = Double.valueOf(previousDayList.get(j).getSetpoint());
                                break;
                            }
                        }
                    }
                }
                break;
            }
        }

        return currentValue.doubleValue();
    }

    /**
     * Checks if the given hour of day is after the setpoint's time.
     *
     * @param hourOfDay
     *            hour of day
     * @param sp
     *            setpoint
     * @return true if the given hour of day is after the setpoint's time
     */
    private boolean isHourInSetpoint(int hourOfDay, Setpoint sp) {
        if (sp == null) {
            return false;
        }

        Calendar c = Calendar.getInstance();
        c.setTime(sp.getTime());
        int setpointHourOfDay = c.get(Calendar.HOUR_OF_DAY);
        int setpointMinute = c.get(Calendar.MINUTE);
        return setpointHourOfDay < hourOfDay || (setpointHourOfDay == hourOfDay && setpointMinute == 0);
    }

    /**
     * Sets the current date.
     *
     * @param date
     *            the current date
     */
    public void setCurrentDate(Calendar date) {
        if (isSingleDay) {
            dayOfWeek = 0;
        } else {
            dayOfWeek = date.get(Calendar.DAY_OF_WEEK) - 1;
        }
    }
}
