/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.service;

import java.io.OutputStream;
import java.util.Date;
import java.util.List;
import java.util.Map;

import com.enerallies.weather.entity.DegreeDayData;
import com.enerallies.weather.entity.DegreeDaySearchCriteria;
import com.enerallies.weather.entity.ExportDirection;
import com.enerallies.weather.entity.WeatherData;
import com.enerallies.weather.entity.WeatherDataSearchCriteria;

/**
 * This service defines the following methods for handling weather data:
 * <ul>
 * <li>retrieve weather data</li>
 * <li>retrieve degree day data</li>
 * <li>export weather data in Excel format to download</li>
 * <li>export degree day data in Excel format to download</li>
 * </ul>
 *
 * <p>
 * <b>Thread Safety:</b> Implementations of this interface must not be thread
 * safe as they will be used in a thread safe manner by the Spring framework.
 * </p>
 *
 * @author Standlove, hanno
 * @version 1.0
 */
public interface WeatherDataService {
    /**
     * Retrieve the weather data according to the passed-in search criteria.
     *
     * @param criteria
     *            the search criteria
     * @return the weather data according to the passed-in search criteria.
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             </ul>
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    public Map<Date, List<WeatherData>> retrieveWeatherData(WeatherDataSearchCriteria criteria)
        throws WeatherDataServiceException;

    /**
     * Retrieve the degree day data according to the passed-in search criteria.
     *
     * @param criteria
     *            the search criteria
     * @return the degree day data according to the passed-in search criteria.
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             </ul>
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    public DegreeDayData retrieveDegreeDayData(DegreeDaySearchCriteria criteria) throws WeatherDataServiceException;

    /**
     * Export weather data in CSV format to output stream.
     *
     *
     * @param weatherData
     *            the data to export.
     * @param direction
     *            the export direction
     * @param stream
     *            the output stream
     * @throws IllegalArgumentException
     *             if any argument is null.
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    public void exportWeatherData(Map<Date, List<WeatherData>> weatherData, ExportDirection direction,
        OutputStream stream) throws WeatherDataServiceException;

    /**
     * Export degree day data in CSV format to output stream.
     *
     * @param degreeDayData
     *            the data to export.
     * @param direction
     *            the export direction
     * @param stream
     *            the output stream
     * @throws IllegalArgumentException
     *             if any argument is null.
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    public void exportDegreeDayData(DegreeDayData degreeDayData, ExportDirection direction, OutputStream stream)
        throws WeatherDataServiceException;
}
