/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.service;

import java.io.BufferedOutputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import com.enerallies.weather.entity.DegreeDay;
import com.enerallies.weather.entity.DegreeDayData;
import com.enerallies.weather.entity.DegreeDaySearchCriteria;
import com.enerallies.weather.entity.DegreeDayType;
import com.enerallies.weather.entity.DegreeType;
import com.enerallies.weather.entity.ExportDirection;
import com.enerallies.weather.entity.Setpoint;
import com.enerallies.weather.entity.SetpointSchedule;
import com.enerallies.weather.entity.WeatherData;
import com.enerallies.weather.entity.WeatherDataSearchCriteria;
import com.enerallies.weather.service.HistoryQueryResultList.HistoryQueryResult;
import com.enerallies.weather.service.Response.Observation;

/**
 * Implementation of the WeatherDataService interface. This provides the
 * following methods for handling weather data:
 * <ul>
 * <li>retrieve weather data</li>
 * <li>retrieve degree day data</li>
 * <li>export weather data in Excel format to download</li>
 * <li>export degree day data in Excel format to download</li>
 * </ul>
 *
 * <p>
 * <b>Thread Safety:</b> This class is mutable and therefore not thread safe but
 * it will not be changed after the Spring injection.
 * </p>
 *
 * @author Standlove, hanno
 * @version 1.0
 */
public class WeatherDataServiceImpl implements WeatherDataService {
    /**
     * Class logger.
     */
    private static final Logger LOGGER = Logger.getLogger(WeatherDataServiceImpl.class);

    /**
     * Base URL.
     */
    private static final String QUERY_BASE = "http://api.wunderground.com/api/{key}/{fct}/q/";

    /**
     * URL to query by ZIP.
     */
    private static final String QUERY_BY_ZIP = QUERY_BASE + "{zip}.xml";

    /**
     * URL to query by State and City.
     */
    private static final String QUERY_BY_STATE = QUERY_BASE + "{state}/{city}.xml";

    /**
     * N/A.
     */
    private static final String N_A = "N/A";

    /**
     * Header title.
     */
    private static final String HEADER_DATE = "Date";

    /**
     * Header title.
     */
    private static final String HEADER_DATE_UTC = "Date (UTC)";

    /**
     * Header title.
     */
    private static final String HEADER_PRECRIP = "Precip (in)";

    /**
     * Header title.
     */
    private static final String HEADER_GUST_SPEED = "Gust Speed (mph)";

    /**
     * Header title.
     */
    private static final String HEADER_WIND_SPEED = "Wind Speed (mph)";

    /**
     * Header title.
     */
    private static final String HEADER_WIND_DIR = "Wind Dir.";

    /**
     * Header title.
     */
    private static final String HEADER_VISIBILITY = "Visibility (mph)";

    /**
     * Header title.
     */
    private static final String HEADER_HUMIDITY = "Humidity (%)";

    /**
     * Header title.
     */
    private static final String HEADER_TEMP = "Temp. (F)";

    /**
     * Header title.
     */
    private static final String HEADER_DEW = "Dew Point (F)";

    /**
     * Header title.
     */
    private static final String HEADER_TIME = "Time (pdt)";

    /**
     * Header for vertical CSV.
     */
    private static final String VERTICAL_HEADER = HEADER_TIME + "," + HEADER_TEMP + "," + HEADER_DEW + ","
        + HEADER_HUMIDITY + "," + HEADER_VISIBILITY + "," + HEADER_WIND_DIR + "," + HEADER_WIND_SPEED + ","
        + HEADER_GUST_SPEED + "," + HEADER_PRECRIP + "," + HEADER_DATE_UTC;

    /**
     * Comma.
     */
    private static final String COMMA = ",";

    /**
     * Setpoints per day.
     */
    private static final int SETPOINTS_PER_DAY = 7;

    /**
     * Setpoints per week.
     */
    private static final int SETPOINTS_PER_WEEK = 7 * SETPOINTS_PER_DAY;

    /**
     * Hours per day.
     */
    private static final int HOURS_PER_DAY = 24;

    /**
     * Fahrenheit to Celcius constant part.
     */
    private static final double F2C_CONST = 32;

    /**
     * Fahrenheit to Celcius factor.
     */
    private static final double F2C_FACTOR = 5.0;

    /**
     * Fahrenheit to Celcius divisor.
     */
    private static final double F2C_DIVISOR = 9.0;

    /**
     * The api key of weather API.
     *
     * You can get one from here:
     * http://www.wunderground.com/weather/api/d/docs?d=index
     *
     * Will be injected by Spring.
     */
    private String weatherDataAPIKey;

    /**
     * Maximum number of API calls made per request.
     *
     * Will be injected by Spring.
     */
    private int apiCallLimit;

    /**
     * Rest template.
     *
     * Injected by Spring.
     */
    @Autowired
    private RestTemplate restTemplate;

    /**
     * Empty constructor for service implementation.
     */
    public WeatherDataServiceImpl() {
        // empty
    }

    /**
     * Retrieve the weather data according to the passed-in search criteria.
     *
     * @param criteria
     *            the search criteria
     * @return the weather data according to the passed-in search criteria.
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             </ul>
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    @Override
    public Map<Date, List<WeatherData>> retrieveWeatherData(WeatherDataSearchCriteria criteria)
        throws WeatherDataServiceException {
        try {
            return retrieveWeatherDataInternal(criteria).toMap();
        } catch (IllegalArgumentException e) {
            LOGGER.error(e.getMessage(), e);
            throw e;
        } catch (WeatherDataServiceException e) {
            LOGGER.error(e.getMessage(), e);
            throw e;
        }
    }

    /**
     * Retrieve the weather data according to the passed-in search criteria.
     *
     * @param criteria
     *            the search criteria
     * @return the weather data according to the passed-in search criteria.
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             </ul>
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    private HistoryQueryResultList retrieveWeatherDataInternal(WeatherDataSearchCriteria criteria)
        throws WeatherDataServiceException {
        validate(criteria);
        return queryAllDays(criteria);
    }

    /**
     * Performs the query for a given date and criteria.
     *
     * @param dateToQuery
     *            the date to query in format YYYYMMDD
     * @param criteria
     *            search criteria
     * @return the response from the query call
     * @throws WeatherDataServiceException
     *             if an error occurs
     */
    private Response doHistoryQuery(String dateToQuery, WeatherDataSearchCriteria criteria)
        throws WeatherDataServiceException {
        Map<String, String> params = new HashMap<String, String>();
        params.put("fct", "history_" + dateToQuery);
        params.put("key", weatherDataAPIKey);

        String query;
        if (criteria.getZipCode() == null || criteria.getZipCode().isEmpty()) {
            query = QUERY_BY_STATE;
            params.put("state", criteria.getState());
            params.put("city", criteria.getCity());
        } else {
            query = QUERY_BY_ZIP;
            params.put("zip", criteria.getZipCode());
        }

        try {
            Response r = restTemplate.getForObject(query, Response.class, params);
            r.setFormattedDate(dateToQuery);
            if (r.getError() != null) {
                throw new WeatherDataServiceException("Query returned " + r.getError().getType() + ": "
                    + r.getError().getDescription());
            }
            return r;
        } catch (RestClientException e) {
            throw new WeatherDataServiceException("Could not query data: " + e.getMessage(), e);
        } catch (WeatherDataServiceException e) {
            throw e;
        } catch (Exception e) {
            throw new WeatherDataServiceException("No data found", e);
        }
    }

    /**
     * This will get a list of all dates that have to be queried in the format
     * YYYYMMDD.
     *
     * @param criteria
     *            the search criteria
     * @return a list of all dates that have to be queried
     * @throws WeatherDataServiceException
     *             if an error occurs
     */
    private HistoryQueryResultList queryAllDays(WeatherDataSearchCriteria criteria) throws WeatherDataServiceException {
        DateFormat df = new SimpleDateFormat("yyyyMMdd");
        Calendar currentDate = Calendar.getInstance();
        currentDate.setTime(criteria.getBeginningDate());

        String end = df.format(criteria.getEndDate());
        String today = df.format(Calendar.getInstance().getTime());

        if (end.compareTo(today) > 0) {
            // no need to query days in the future
            end = today;
        }

        String currentFormatted = df.format(currentDate.getTime());
        List<Pair<Calendar, String>> formattedDates = new ArrayList<Pair<Calendar, String>>();

        HistoryQueryResultList result = new HistoryQueryResultList(criteria);

        while (currentFormatted.compareTo(end) <= 0) {
            formattedDates.add(new ImmutablePair<Calendar, String>((Calendar) currentDate.clone(), currentFormatted));

            // move on to the next day
            currentDate.add(Calendar.DATE, 1);
            currentFormatted = df.format(currentDate.getTime());
        }

        LOGGER.debug("Number of calls necessary: " + formattedDates.size());
        if (formattedDates.size() > apiCallLimit) {
            throw new WeatherDataServiceException(MessageFormat.format(
                "Number of API calls would be {0} which exceeds allowed maximum of {1}. Please lower the date range.",
                Integer.valueOf(formattedDates.size()), Integer.valueOf(apiCallLimit)));
        }

        for (Pair<Calendar, String> e : formattedDates) {
            result.add(e.getKey(), doHistoryQuery(e.getValue(), criteria));
        }

        result.interpolate();

        return result;
    }

    /**
     * Validates the search criteria.
     *
     * @param criteria
     *            criteria
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             </ul>
     */
    private void validate(WeatherDataSearchCriteria criteria) {
        if (criteria == null) {
            throw new IllegalArgumentException("Criteria must not be null");
        }
        if (criteria.getBeginningDate() == null) {
            throw new IllegalArgumentException("Begin date must not be null");
        }
        if (criteria.getEndDate() == null) {
            throw new IllegalArgumentException("End date must not be null");
        }
        if (criteria.getBeginningDate().after(criteria.getEndDate())) {
            throw new IllegalArgumentException("Begin date must not be after end date");
        }
        if (criteria.getZipCode() == null && (criteria.getCity() == null || criteria.getCity().isEmpty())) {
            throw new IllegalArgumentException("City must not be null or empty");
        }
        if (criteria.getZipCode() == null && (criteria.getState() == null || criteria.getState().isEmpty())) {
            throw new IllegalArgumentException("State must not be null or empty");
        }

        Calendar today = Calendar.getInstance();
        today.set(Calendar.HOUR, 0);
        today.set(Calendar.MINUTE, 0);
        today.set(Calendar.SECOND, 0);
        today.set(Calendar.MILLISECOND, 0);
        today.add(Calendar.HOUR, 24);

        if (criteria.getBeginningDate().after(today.getTime())) {
            throw new IllegalArgumentException("Begin date must not be in the future");
        }
    }

    /**
     * Validates the search criteria.
     *
     * @param criteria
     *            criteria
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             <li>degreeDayType is null</li>
     *             <li>setpoints is null</li>
     *             <li>degreeType is null</li>
     *             <li>setpointSchedule is null</li>
     *             <li>if setpointSchedule is SINGLE_DAY and setpoints.size is
     *             not 7</li>
     *             <li>if setpointSchedule is SEVEN_DAYS and setpoints.size is
     *             not 7x7</li>
     *             </ul>
     */
    private void validate(DegreeDaySearchCriteria criteria) {
        validate((WeatherDataSearchCriteria) criteria);

        if (criteria.getDegreeDayType() == null) {
            throw new IllegalArgumentException("Degree day type must not be null");
        }
        if (criteria.getDegreeType() == null) {
            throw new IllegalArgumentException("Degree type must not be null");
        }
        if (criteria.getSetpoints() == null) {
            throw new IllegalArgumentException("Setpoints must not be null");
        }
        if (criteria.getSetpointSchedule() == null) {
            throw new IllegalArgumentException("Setpoint schedule must not be null");
        }
        if (criteria.getSetpointSchedule() == SetpointSchedule.SINGLE_DAY
            && criteria.getSetpoints().size() != SETPOINTS_PER_DAY) {
            throw new IllegalArgumentException("For schedule SINGLE_DAY there should be 7 setpoints");
        }
        if (criteria.getSetpointSchedule() == SetpointSchedule.SEVEN_DAYS
            && criteria.getSetpoints().size() != SETPOINTS_PER_WEEK) {
            throw new IllegalArgumentException("For schedule SEVEN_DAYS there should be 49 (7x7) setpoints");
        }
        for (Setpoint sp : criteria.getSetpoints()) {
            if (sp == null) {
                continue;
            }
            if (sp.getTime() == null) {
                throw new IllegalArgumentException("Setpoint is missing date value");
            }
        }

        Calendar today = Calendar.getInstance();
        today.set(Calendar.HOUR, 0);
        today.set(Calendar.MINUTE, 0);
        today.set(Calendar.SECOND, 0);
        today.set(Calendar.MILLISECOND, 0);

        if (!criteria.getEndDate().before(today.getTime())) {
            throw new IllegalArgumentException("Degree Day data may only be calculated for past dates");
        }
    }

    /**
     * Retrieve the degree day data according to the passed-in search criteria.
     *
     * @param criteria
     *            the search criteria
     * @return the degree day data according to the passed-in search criteria.
     * @throws IllegalArgumentException
     *             if any of the following is true:
     *             <ul>
     *             <li>criteria is null</li>
     *             <li>either beginningDate or endDate is null</li>
     *             <li>beginningDate is after endDate</li>
     *             <li>if beginningDate or endDate are in the future</li>
     *             <li>zipCode is null and city is null or empty</li>
     *             <li>zipCode is null and state is null or empty</li>
     *             <li>degreeDayType is null</li>
     *             <li>setpoints is null</li>
     *             <li>degreeType is null</li>
     *             <li>setpointSchedule is null</li>
     *             <li>if setpointSchedule is SINGLE_DAY and setpoints.size is
     *             not 7</li>
     *             <li>if setpointSchedule is SEVEN_DAYS and setpoints.size is
     *             not 7x7</li>
     *             </ul>
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    @Override
    public DegreeDayData retrieveDegreeDayData(DegreeDaySearchCriteria criteria) throws WeatherDataServiceException {
        try {
            validate(criteria);
            HistoryQueryResultList wd = retrieveWeatherDataInternal(criteria);

            DegreeDayData degreeDayData = new DegreeDayData();
            degreeDayData.setCriteria(criteria);
            degreeDayData.setDegreeDays(new ArrayList<DegreeDay>());
            degreeDayData.setDegreeType(criteria.getDegreeType());
            degreeDayData.setDegreeDayType(criteria.getDegreeDayType());
            degreeDayData.setTotalDD(0);

            SetpointList spList = new SetpointList(criteria);

            DateFormat df = new SimpleDateFormat("MM/dd/yyyy");
            DecimalFormat decf = new DecimalFormat("#.00");

            for (HistoryQueryResult res : wd.getData()) {
                Calendar currentDate = res.getDate();
                Observation[] observations = res.getObservations();

                DegreeDay degreeDay = new DegreeDay();
                degreeDay.setDate(currentDate.getTime());

                spList.setCurrentDate(currentDate);
                double degreeDayValue = 0;
                for (int hourOfDay = 0; hourOfDay < observations.length; hourOfDay++) {
                    double outsideTemp = convertFahrenheit(observations[hourOfDay].getTemperature(),
                        criteria.getDegreeType());

                    double setpointValue = spList.getValue(hourOfDay);

                    LOGGER.debug("Outside temperature on " + df.format(currentDate.getTime()) + " at "
                        + (hourOfDay < 10 ? "0" : "") + hourOfDay + ":00 is " + decf.format(outsideTemp) + " "
                        + criteria.getDegreeType() + " with setpoint " + decf.format(setpointValue));

                    double diff = 0;
                    if (criteria.getDegreeDayType() == DegreeDayType.COOLING) {
                        diff = outsideTemp - setpointValue;
                    } else {
                        diff = setpointValue - outsideTemp;
                    }

                    if (diff > 0) {
                        degreeDayValue += diff / HOURS_PER_DAY;
                    }
                }

                degreeDay.setDegreeDay(degreeDayValue);
                degreeDayData.getDegreeDays().add(degreeDay);
                degreeDayData.setTotalDD(degreeDayData.getTotalDD() + degreeDayValue);
            }

            return degreeDayData;
        } catch (IllegalArgumentException e) {
            LOGGER.error(e.getMessage(), e);
            throw e;
        } catch (WeatherDataServiceException e) {
            LOGGER.error(e.getMessage(), e);
            throw e;
        }
    }

    /**
     * Convert the given value from Fahrenheit to the given type.
     *
     * @param valueF
     *            value in Fahrenheit
     * @param type
     *            type to be converted to
     * @return converted value
     */
    private double convertFahrenheit(Double valueF, DegreeType type) {
        if (valueF == null) {
            return 0;
        }
        if (type == DegreeType.FAHRENHEIT) {
            return valueF.doubleValue();
        }
        return (valueF.doubleValue() - F2C_CONST) * F2C_FACTOR / F2C_DIVISOR;
    }

    /**
     * Export weather data in CSV format to output stream.
     *
     *
     * @param weatherData
     *            the data to export.
     * @param direction
     *            the export direction
     * @param stream
     *            the output stream
     * @throws IllegalArgumentException
     *             if any argument is null.
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    @Override
    public void exportWeatherData(Map<Date, List<WeatherData>> weatherData, ExportDirection direction,
        OutputStream stream) throws WeatherDataServiceException {
        if (weatherData == null) {
            throw new IllegalArgumentException("Weather data must not be null");
        }
        if (direction == null) {
            throw new IllegalArgumentException("Direction must not be null");
        }
        if (stream == null) {
            throw new IllegalArgumentException("Stream must not be null");
        }

        PrintWriter pw = null;
        try {
            WeatherData wd = weatherData.entrySet().iterator().next().getValue().get(0);
            pw = new PrintWriter(new BufferedOutputStream(stream));
            printFirstLine(pw, wd.getCriteria());

            List<Date> dates = new ArrayList<Date>(weatherData.keySet());
            Collections.sort(dates);

            SimpleDateFormat timeFormat = new SimpleDateFormat("yyyy/MM/dd hh:mm aa");
            timeFormat.setTimeZone(TimeZone.getTimeZone(wd.getTimeZone()));

            SimpleDateFormat utcFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm");
            utcFormat.setTimeZone(TimeZone.getTimeZone("UTC"));

            if (direction == ExportDirection.VERTICAL) {
                exportVertical(pw, dates, weatherData, timeFormat, utcFormat);
            } else {
                exportHorizontal(pw, dates, weatherData, timeFormat, utcFormat);
            }
        } finally {
            close(pw);
        }
    }

    /**
     * Export horizontal.
     *
     * @param pw
     *            print writer
     * @param dates
     *            dates list
     * @param weatherData
     *            weather data
     * @param timeFormat
     *            time format
     * @param utcFormat
     *            UTC date format
     */
    private void exportHorizontal(PrintWriter pw, List<Date> dates, Map<Date, List<WeatherData>> weatherData,
        SimpleDateFormat timeFormat, SimpleDateFormat utcFormat) {

        write(pw, HEADER_TIME);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, timeFormat.format(w.getTime()), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_TEMP);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getTemperature(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_DEW);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getDewPoint(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_HUMIDITY);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getHumidity(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_VISIBILITY);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getVisibility(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_WIND_DIR);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getWindDirection(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_WIND_SPEED);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getWindSpeed(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_GUST_SPEED);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getGustSpeed(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_PRECRIP);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, w.getPrecip(), COMMA);
            }
        }
        pw.println();

        write(pw, HEADER_DATE_UTC);
        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);
            for (WeatherData w : wd) {
                write(pw, utcFormat.format(w.getUtcDate()), COMMA);
            }
        }
        pw.println();
    }

    /**
     * Export vertical.
     *
     * @param pw
     *            print writer
     * @param dates
     *            dates list
     * @param weatherData
     *            weather data
     * @param timeFormat
     *            time format
     * @param utcFormat
     *            UTC date format
     */
    private void exportVertical(PrintWriter pw, List<Date> dates, Map<Date, List<WeatherData>> weatherData,
        SimpleDateFormat timeFormat, SimpleDateFormat utcFormat) {
        pw.println(VERTICAL_HEADER);

        for (Date d : dates) {
            List<WeatherData> wd = weatherData.get(d);

            for (WeatherData w : wd) {
                write(pw, timeFormat.format(w.getTime()));
                write(pw, w.getTemperature(), COMMA);
                write(pw, w.getDewPoint(), COMMA);
                write(pw, w.getHumidity(), COMMA);
                write(pw, w.getVisibility(), COMMA);
                write(pw, w.getWindDirection(), COMMA);
                write(pw, w.getWindSpeed(), COMMA);
                write(pw, w.getGustSpeed(), COMMA);
                write(pw, w.getPrecip(), COMMA);
                write(pw, utcFormat.format(w.getUtcDate()), COMMA);
                pw.println();
            }
        }
    }

    /**
     * Write a value to output.
     *
     * @param w
     *            print writer
     * @param d
     *            value
     */
    private void write(PrintWriter w, String d) {
        write(w, d, "");
    }

    /**
     * Write a value to output.
     *
     * @param w
     *            print writer
     * @param d
     *            value
     * @param separator
     *            separator before value
     */
    private void write(PrintWriter w, String d, String separator) {
        w.print(separator);
        if (d == null) {
            w.print(N_A);
        } else {
            w.print(d);
        }
    }

    /**
     * Write a value to output.
     *
     * @param w
     *            print writer
     * @param d
     *            value
     * @param separator
     *            separator before value
     */
    private void write(PrintWriter w, Double d, String separator) {
        w.print(separator);
        if (d != null) {
            w.format(Locale.US, "%1$.2f", d);
        } else {
            w.print("");
        }
    }

    /**
     * Write a value to output.
     *
     * @param w
     *            print writer
     * @param d
     *            value
     * @param separator
     *            separator before value
     */
    private void write(PrintWriter w, Integer d, String separator) {
        w.print(separator);
        if (d != null) {
            w.print(d);
        } else {
            w.print("");
        }
    }

    /**
     * Close silently.
     *
     * @param w
     *            writer to be closed.
     */
    private void close(Closeable w) {
        if (w != null) {
            try {
                w.close();
            } catch (IOException e) {
                LOGGER.error("Error closing stream: " + e.getMessage(), e);
            }
        }
    }

    /**
     * Export degree day data in CSV format to output stream.
     *
     * @param degreeDayData
     *            the data to export.
     * @param direction
     *            the export direction
     * @param stream
     *            the output stream
     * @throws IllegalArgumentException
     *             if any argument is null.
     * @throws WeatherDataServiceException
     *             if any other error occurs
     */
    @Override
    public void exportDegreeDayData(DegreeDayData degreeDayData, ExportDirection direction, OutputStream stream)
        throws WeatherDataServiceException {
        if (degreeDayData == null) {
            throw new IllegalArgumentException("Degree day data must not be null");
        }
        if (direction == null) {
            throw new IllegalArgumentException("Direction must not be null");
        }
        if (stream == null) {
            throw new IllegalArgumentException("Stream must not be null");
        }

        PrintWriter pw = null;
        try {
            pw = new PrintWriter(new BufferedOutputStream(stream));

            // HEADER
            printFirstLine(pw, degreeDayData.getCriteria());
            write(pw, "Description:");
            write(pw, MessageFormat.format("{0}-based {1} degree days",
                degreeDayData.getDegreeType() == DegreeType.FAHRENHEIT ? "Fahrenheit" : "Celcius",
                degreeDayData.getDegreeDayType() == DegreeDayType.COOLING ? "cooling" : "heating"), COMMA);
            pw.println();

            write(pw, "Station:");
            write(pw, degreeDayData.getStation(), COMMA);
            pw.println();

            write(pw, "Station ID:");
            write(pw, degreeDayData.getStationId(), COMMA);
            pw.println();
            pw.println();

            // CONTENT
            SimpleDateFormat timeFormat = new SimpleDateFormat("MM/dd/yyyy");
            if (direction == ExportDirection.VERTICAL) {
                exportVertical(pw, degreeDayData, timeFormat);
            } else {
                exportHorizontal(pw, degreeDayData, timeFormat);
            }

            // TOTAL
            pw.println();
            write(pw, "Total DD:");
            write(pw, Double.valueOf(degreeDayData.getTotalDD()), COMMA);
        } finally {
            close(pw);
        }
    }

    /**
     * Print the first line.
     *
     * @param pw
     *            print writer
     * @param criteria
     *            search criteria
     */
    private void printFirstLine(PrintWriter pw, WeatherDataSearchCriteria criteria) {
        pw.print('"');
        if (criteria.getZipCode() != null && !criteria.getZipCode().isEmpty()) {
            pw.print(criteria.getZipCode());
        } else {
            pw.print(criteria.getCity() + ", " + criteria.getState() + ".");
        }

        SimpleDateFormat timeFormat = new SimpleDateFormat("MMMM dd, yyyy", Locale.US);
        pw.print(" " + timeFormat.format(criteria.getBeginningDate()));
        if (criteria.getEndDate().after(Calendar.getInstance().getTime())) {
            pw.print(" - " + timeFormat.format(Calendar.getInstance().getTime()));
        } else {
            pw.print(" - " + timeFormat.format(criteria.getEndDate()));
        }
        pw.println('"');
    }

    /**
     * Export vertical.
     *
     * @param pw
     *            print writer
     * @param data
     *            data
     * @param timeFormat
     *            time format
     */
    private void exportVertical(PrintWriter pw, DegreeDayData data, SimpleDateFormat timeFormat) {
        pw.print(HEADER_DATE);
        pw.print(COMMA);
        printDDHeader(pw, data);
        pw.println();

        for (DegreeDay d : data.getDegreeDays()) {
            write(pw, timeFormat.format(d.getDate()));
            write(pw, Double.valueOf(d.getDegreeDay()), COMMA);
            pw.println();
        }
    }

    /**
     * Writes HDD or CDD.
     *
     * @param pw
     *            print writer
     * @param data
     *            data
     */
    private void printDDHeader(PrintWriter pw, DegreeDayData data) {
        if (data.getDegreeDayType() == DegreeDayType.HEATING) {
            pw.print("HDD");
        } else {
            pw.print("CDD");
        }
    }

    /**
     * Export horizontal.
     *
     * @param pw
     *            print writer
     * @param data
     *            data
     * @param timeFormat
     *            time format
     */
    private void exportHorizontal(PrintWriter pw, DegreeDayData data, SimpleDateFormat timeFormat) {
        write(pw, HEADER_DATE);
        for (DegreeDay d : data.getDegreeDays()) {
            write(pw, timeFormat.format(d.getDate()), COMMA);
        }
        pw.println();

        printDDHeader(pw, data);
        for (DegreeDay d : data.getDegreeDays()) {
            write(pw, Double.valueOf(d.getDegreeDay()), COMMA);
        }
        pw.println();
    }

    /**
     * Sets the weatherDataAPIKey.
     *
     * @param weatherDataAPIKey
     *            the new weatherDataAPIKey to set
     */
    public void setWeatherDataAPIKey(String weatherDataAPIKey) {
        this.weatherDataAPIKey = weatherDataAPIKey;
    }

    /**
     * Sets the restTemplate.
     *
     * @param restTemplate
     *            the new restTemplate to set
     */
    void setRestTemplate(RestTemplate restTemplate) {
        this.restTemplate = restTemplate;
    }

    /**
     * Returns the apiCallLimit.
     *
     * @return the apiCallLimit
     */
    public int getApiCallLimit() {
        return apiCallLimit;
    }

    /**
     * Sets the apiCallLimit.
     *
     * @param apiCallLimit
     *            the new apiCallLimit to set
     */
    public void setApiCallLimit(int apiCallLimit) {
        this.apiCallLimit = apiCallLimit;
    }
}
