/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.service;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

import junit.framework.Assert;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import com.enerallies.weather.entity.DegreeDaySearchCriteria;
import com.enerallies.weather.entity.Setpoint;
import com.enerallies.weather.entity.SetpointSchedule;

/**
 * Tests for SetpointList class.
 *
 * @author hanno
 * @version 1.0
 */
public class SetpointListTest {

    /**
     * Days of the week.
     */
    private static final int[] DAYS = new int[] {Calendar.MONDAY, Calendar.TUESDAY, Calendar.WEDNESDAY,
        Calendar.THURSDAY, Calendar.FRIDAY, Calendar.SATURDAY, Calendar.SUNDAY };

    /**
     * Delta.
     */
    private static final double DELTA = 0.00001;

    /**
     * Calendar mock.
     */
    @Mock
    private Calendar c;

    /**
     * Test set up.
     *
     * @throws Exception
     *             to JUnit
     */
    @Before
    public void setUp() throws Exception {
        MockitoAnnotations.initMocks(this);
    }

    /**
     * Creates a new setpoint.
     *
     * @param date
     *            date
     * @param value
     *            value
     * @return new setpoint
     * @throws ParseException
     *             to JUnit
     */
    private Setpoint createSetpoint(String date, double value) throws ParseException {
        SimpleDateFormat df = new SimpleDateFormat("HH:mm");
        Setpoint sp = new Setpoint();
        sp.setSetpoint(value);
        sp.setTime(df.parse(date));
        return sp;
    }

    /**
     * Test clean up.
     */
    @After
    public void tearDown() {
        this.c = null;
    }

    /**
     * Creates criteria.
     *
     * @param s
     *            schedule
     * @return criteria
     */
    private DegreeDaySearchCriteria createDegreeDaySearchCriteria(SetpointSchedule s) {
        DegreeDaySearchCriteria crit = new DegreeDaySearchCriteria();
        crit.setSetpointSchedule(s);
        crit.setSetpoints(new ArrayList<Setpoint>());

        for (int i = 0; i < (s == SetpointSchedule.SINGLE_DAY ? 7 : 49); i++) {
            crit.getSetpoints().add(null);
        }

        return crit;
    }

    /**
     * Test for a single day with multiple setpoints.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test
    public void singleDayMultipleSetpoints() throws Exception {
        DegreeDaySearchCriteria critSingleDay = createDegreeDaySearchCriteria(SetpointSchedule.SINGLE_DAY);

        critSingleDay.getSetpoints().set(0, createSetpoint("07:30", 1));
        critSingleDay.getSetpoints().set(1, createSetpoint("14:00", 2));
        critSingleDay.getSetpoints().set(2, createSetpoint("22:01", 3));

        SetpointList spList = new SetpointList(critSingleDay);
        for (int day : DAYS) {
            assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3 }, day,
                spList);
        }
    }

    /**
     * Test for a single day with one setpoint.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test
    public void singleDayOneSetpoint() throws Exception {
        DegreeDaySearchCriteria critSingleDay = createDegreeDaySearchCriteria(SetpointSchedule.SINGLE_DAY);

        critSingleDay.getSetpoints().set(0, createSetpoint("08:00", 1));

        SetpointList spList = new SetpointList(critSingleDay);
        for (int day : DAYS) {
            assertDay(new double[] {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }, day,
                spList);
        }
    }

    /**
     * Test for a single day with no setpoint.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void singleDayNoSetpoint() throws Exception {
        DegreeDaySearchCriteria critSingleDay = createDegreeDaySearchCriteria(SetpointSchedule.SINGLE_DAY);

        new SetpointList(critSingleDay);
    }

    /**
     * Test for a 7 days with multiple setpoints.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test
    public void sevenDaysMultipleSetpointsSunday() throws Exception {
        DegreeDaySearchCriteria critSingleDay = createDegreeDaySearchCriteria(SetpointSchedule.SEVEN_DAYS);

        critSingleDay.getSetpoints().set(0, createSetpoint("08:00", 1));
        critSingleDay.getSetpoints().set(1, createSetpoint("14:00", 2));
        critSingleDay.getSetpoints().set(2, createSetpoint("22:00", 3));

        SetpointList spList = new SetpointList(critSingleDay);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3 },
            Calendar.SUNDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.MONDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.TUESDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.WEDNESDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.THURSDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.FRIDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.SATURDAY, spList);
    }

    /**
     * Test for a 7 days with single setpoints.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test
    public void sevenDaysSingleSetpointSunday() throws Exception {
        DegreeDaySearchCriteria critSingleDay = createDegreeDaySearchCriteria(SetpointSchedule.SEVEN_DAYS);

        critSingleDay.getSetpoints().set(0, createSetpoint("08:00", 3));

        SetpointList spList = new SetpointList(critSingleDay);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.SUNDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.MONDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.TUESDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.WEDNESDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.THURSDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.FRIDAY, spList);
        assertDay(new double[] {3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
            Calendar.SATURDAY, spList);
    }

    /**
     * Assert values for a day.
     *
     * @param values
     *            values
     * @param day
     *            the day
     * @param spList
     *            setpoint list
     */
    private void assertDay(double[] values, int day, SetpointList spList) {
        Assert.assertEquals("number of values per day", 24, values.length);
        Mockito.when(Integer.valueOf(c.get(Calendar.DAY_OF_WEEK))).thenReturn(Integer.valueOf(day));

        spList.setCurrentDate(c);

        for (int i = 0; i < values.length; i++) {
            Assert.assertEquals("value " + i, values[i], spList.getValue(i), DELTA);
        }
    }
}
