/*
 * Copyright (C) 2012 TopCoder Inc., All Rights Reserved.
 */
package com.enerallies.weather.service;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXB;

import junit.framework.Assert;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import com.enerallies.weather.entity.DegreeDaySearchCriteria;
import com.enerallies.weather.entity.DegreeDayType;
import com.enerallies.weather.entity.DegreeType;
import com.enerallies.weather.entity.Setpoint;
import com.enerallies.weather.entity.SetpointSchedule;
import com.enerallies.weather.entity.WeatherData;
import com.enerallies.weather.entity.WeatherDataSearchCriteria;

/**
 * Tests for WeatherDataServiceImpl class.
 *
 * @author hanno
 * @version 1.0
 */
@SuppressWarnings({"unchecked", "rawtypes" })
public class WeatherDataServiceImplTest {
    /**
     * Service to be tested.
     */
    private WeatherDataServiceImpl impl;

    /**
     * Weather search criteria.
     */
    private WeatherDataSearchCriteria weatherSearchCriteria;

    /**
     * Degree day search criteria.
     */
    private DegreeDaySearchCriteria degreeDaySearchCriteria;

    /**
     * RetstTemplate mock.
     */
    @Mock
    private RestTemplate restTemplate;

    /**
     * Response.
     */
    private Response responseOK;

    /**
     * NOK response.
     */
    private Response responseNOK;

    /**
     * OK response with only a single value.
     */
    private Response responseOKSingleValue;

    /**
     * OK response empty.
     */
    private Response responseOKEmpty;

    /**
     * Test set up.
     *
     * @throws Exception
     *             to JUnit
     */
    @Before
    public void setUp() throws Exception {
        MockitoAnnotations.initMocks(this);

        this.impl = new WeatherDataServiceImpl();
        this.impl.setWeatherDataAPIKey("123");
        this.impl.setApiCallLimit(123);
        this.impl.setRestTemplate(this.restTemplate);

        this.weatherSearchCriteria = new WeatherDataSearchCriteria();
        initCriteria(this.weatherSearchCriteria);

        this.degreeDaySearchCriteria = new DegreeDaySearchCriteria();
        initCriteria(this.degreeDaySearchCriteria);

        this.responseOK = JAXB.unmarshal(getClass().getResourceAsStream("/history_response_OK.xml"), Response.class);
        this.responseNOK = JAXB.unmarshal(getClass().getResourceAsStream("/history_response_NOK.xml"), Response.class);
        this.responseOKSingleValue = JAXB.unmarshal(
            getClass().getResourceAsStream("/history_response_OK_singleValue.xml"), Response.class);
        this.responseOKEmpty = JAXB.unmarshal(getClass().getResourceAsStream("/history_response_Empty.xml"),
            Response.class);

        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenReturn(this.responseOK);
    }

    /**
     * Initialize test values.
     *
     * @param crit
     *            criteria.
     */
    private void initCriteria(WeatherDataSearchCriteria crit) {
        crit.setBeginningDate(new Date(10));
        crit.setCity("city");
        crit.setEndDate(new Date(20));
        crit.setState("state");
        crit.setZipCode("zip");
    }

    /**
     * Initialize test values.
     *
     * @param crit
     *            criteria.
     */
    private void initCriteria(DegreeDaySearchCriteria crit) {
        initCriteria((WeatherDataSearchCriteria) crit);
        crit.setDegreeDayType(DegreeDayType.COOLING);
        crit.setDegreeType(DegreeType.CELSIUS);
        crit.setSetpointSchedule(SetpointSchedule.SINGLE_DAY);

        List<Setpoint> sp = new ArrayList<Setpoint>();
        for (int i = 0; i < 7; i++) {
            Setpoint s = new Setpoint();
            s.setSetpoint(i);
            s.setTime(new Date(i));
            sp.add(s);
        }
        crit.setSetpoints(sp);
    }

    /**
     * Test clean up.
     */
    @After
    public void tearDown() {
        this.impl = null;
        this.weatherSearchCriteria = null;
        this.degreeDaySearchCriteria = null;
        this.restTemplate = null;
        this.responseNOK = null;
        this.responseOK = null;
    }

    /**
     * retrieveWeatherData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveWeatherDataFailureNullCrit() throws Exception {
        this.impl.retrieveWeatherData(null);
    }

    /**
     * retrieveWeatherData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveWeatherDataFailureNullBegin() throws Exception {
        this.weatherSearchCriteria.setBeginningDate(null);
        this.impl.retrieveWeatherData(this.weatherSearchCriteria);
    }

    /**
     * retrieveWeatherData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveWeatherDataFailureNullEnd() throws Exception {
        this.weatherSearchCriteria.setEndDate(null);
        this.impl.retrieveWeatherData(this.weatherSearchCriteria);
    }

    /**
     * retrieveWeatherData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveWeatherDataFailureBeginAfterEnd() throws Exception {
        this.weatherSearchCriteria.setBeginningDate(new Date(1000));
        this.impl.retrieveWeatherData(this.weatherSearchCriteria);
    }

    /**
     * Tests retrieveWeatherData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = WeatherDataServiceException.class)
    public void retrieveWeatherDataClientException() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenThrow(new RestClientException("test"));
        this.impl.retrieveWeatherData(this.weatherSearchCriteria);
    }

    /**
     * Tests retrieveWeatherData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = WeatherDataServiceException.class)
    public void retrieveWeatherDataNokReponse() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenReturn(this.responseNOK);
        this.impl.retrieveWeatherData(this.weatherSearchCriteria);
    }

    /**
     * Tests retrieveWeatherData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = WeatherDataServiceException.class)
    public void retrieveWeatherDataEmptyReponse() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenReturn(this.responseOKEmpty);
        this.impl.retrieveWeatherData(this.weatherSearchCriteria);
    }

    /**
     * Tests retrieveWeatherData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test
    public void retrieveWeatherData() throws Exception {
        Map<Date, List<WeatherData>> data = this.impl.retrieveWeatherData(this.weatherSearchCriteria);

        Assert.assertEquals("size", 1, data.size());
        List<WeatherData> wd = data.get(data.keySet().iterator().next());

        Assert.assertEquals("values per day", 24, wd.size());
    }

    /**
     * Tests retrieveWeatherData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test
    public void retrieveWeatherDataSingleValue() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenReturn(this.responseOKSingleValue);
        Map<Date, List<WeatherData>> data = this.impl.retrieveWeatherData(this.weatherSearchCriteria);

        Assert.assertEquals("size", 1, data.size());
        List<WeatherData> wd = data.get(data.keySet().iterator().next());

        Assert.assertEquals("values per day", 24, wd.size());
        // all should have same values
        for (int i = 0; i < 24; i++) {
            Assert.assertEquals("Dew point", new Double(35.6), wd.get(i).getDewPoint());
            Assert.assertEquals("gust speed", null, wd.get(i).getGustSpeed());
            Assert.assertEquals("humidity", new Integer(81), wd.get(i).getHumidity());
            Assert.assertEquals("precip", null, wd.get(i).getPrecip());
            Assert.assertEquals("temperature", new Double(41.0), wd.get(i).getTemperature());
            Assert.assertEquals("visibility", new Double(24.1), wd.get(i).getVisibility());
            Assert.assertEquals("wind direction", "WSW", wd.get(i).getWindDirection());
            Assert.assertEquals("wind speed", new Double(11.1), wd.get(i).getWindSpeed());
        }
    }

    /**
     * retrieveDegreeDayData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveDegreeDayDataFailureNullCrit() throws Exception {
        this.impl.retrieveDegreeDayData(null);
    }

    /**
     * retrieveDegreeDayData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveDegreeDayDataFailureNullBegin() throws Exception {
        this.degreeDaySearchCriteria.setBeginningDate(null);
        this.impl.retrieveDegreeDayData(this.degreeDaySearchCriteria);
    }

    /**
     * retrieveDegreeDayData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveDegreeDayDataFailureNullEnd() throws Exception {
        this.degreeDaySearchCriteria.setEndDate(null);
        this.impl.retrieveDegreeDayData(this.degreeDaySearchCriteria);
    }

    /**
     * retrieveDegreeDayData failure test.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = IllegalArgumentException.class)
    public void retrieveDegreeDayDataFailureBeginAfterEnd() throws Exception {
        this.degreeDaySearchCriteria.setBeginningDate(new Date(1000));
        this.impl.retrieveDegreeDayData(this.degreeDaySearchCriteria);
    }

    /**
     * Tests retrieveDegreeDayData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = WeatherDataServiceException.class)
    public void retrieveDegreeDayDataClientException() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenThrow(new RestClientException("test"));
        this.impl.retrieveDegreeDayData(this.degreeDaySearchCriteria);
    }

    /**
     * Tests retrieveDegreeDayData.
     *
     * @throws Exception
     *             to JUnit
     */
    @Test(expected = WeatherDataServiceException.class)
    public void retrieveDegreeDayDataNokReponse() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenReturn(this.responseNOK);
        this.impl.retrieveDegreeDayData(this.degreeDaySearchCriteria);
    }

    /**
     * Tests retrieveDegreeDayData.
     *
     * @throws Exception
     *             to JUnit
     */

    @Test(expected = WeatherDataServiceException.class)
    public void retrieveDegreeDayDataEmptyReponse() throws Exception {
        Mockito.when(
            this.restTemplate.getForObject(Mockito.anyString(), (Class) Mockito.any(), (Map<String, ?>) Mockito.any()))
            .thenReturn(this.responseOKEmpty);
        this.impl.retrieveDegreeDayData(this.degreeDaySearchCriteria);
    }
}
